/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _SPLAT_DOT_PRODUCT3_H_
#define _SPLAT_DOT_PRODUCT3_H_	1

#ifdef __SPU__
#include <spu_intrinsics.h>
#else /* not __SPU__ */
#include <altivec.h>
#endif

/*
 * FUNCTION
 * 	vector float _splat_dot_product3(vector float v1, vector float v2)
 * 
 * DESCRIPTION
 * 	_vector_dot_product3 computes the dot product of the two input 
 *	vectors, v1 dot v2, and results the result replicated (splated)
 *	across a 128-bit floating-point vector.. The inputs v1 and v2 
 *	are 4 component vectors in which only the first 3 (3 most significant)
 *	channels contribute to the dot product computation. 
 *           _______________________________
 *          |___X___|___Y___|___Z___|___W___|
 *        
 *	result = (v1.x * v2.x) +  (v1.y * v2.y) + (v1.z * v2.z);
 */

static __inline vector float _splat_dot_product3(vector float v1, vector float v2)
{
  vector float product;
  vector float y, z;
  vector float result;


#ifdef __SPU__
  product = spu_mul(v1, v2);

  y = spu_rlqwbyte(product, 4);
  z = spu_rlqwbyte(product, 8);
  result = spu_add(product, y);
  result = spu_add(result, z);

  result = spu_shuffle(result, result, (vector unsigned char)(spu_splats((unsigned int)0x00010203)));

#else
  vector float x;

  product = vec_madd(v1, v2, ((vector float) {0.0f,0.0f,0.0f,0.0f}));

  x = vec_splat(product, 0);
  y = vec_splat(product, 1);
  z = vec_splat(product, 2);
  result = vec_add(x, y);
  result = vec_add(result, z);
#endif

  return (result);
}


#endif /* _SPLAT_DOT_PRODUCT3_H_ */
