/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#ifndef _MPM_CMPEQ2_H_
#define _MPM_CMPEQ2_H_	1

#include <spu_intrinsics.h>

/* mpm_cmpeq2
 * ----------
 * Generic routine that compares two large numbers pointed to by <a> and <b> 
 * of <asize> and <bsize> quadwords respectively. If the two numbers are equal,
 * then all 1;s is returned, else 0 is returned.
 *
 */
static __inline unsigned int _mpm_cmpeq2(const vector unsigned int *a, int asize, const vector unsigned int *b, int bsize)
{
  int i, j;
  int delta;
  vector unsigned int equal, ors;

  equal = spu_splats((unsigned int)-1);

  delta = asize - bsize;
  if (delta > 0) {
    ors = spu_splats((unsigned int)0);
    for (i=0; i<delta; i++) {
      ors = spu_or(ors, a[i]);
    }
    equal = spu_and(equal, spu_cmpeq(ors, 0));
    for (; i<asize; i++) {
      j = i - delta;
      equal = spu_and(equal, spu_cmpeq(a[i], b[j]));
    }
  } else {
    delta = -delta;

    ors = spu_splats((unsigned int)0);
    for (i=0; i<delta; i++) {
      ors = spu_or(ors, b[i]);
    }
    equal = spu_and(equal, spu_cmpeq(ors, 0));
    for (; i<bsize; i++) {
      j = i - delta;
      equal = spu_and(equal, spu_cmpeq(a[j], b[i]));
    }
  }
  return (spu_extract(spu_cmpeq(spu_gather(equal), 15), 0));
}

#endif /* _MPM_CMPEQ2_H_ */
