/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#ifndef _MPM_MONT_MOD_EXP3_H_
#define _MPM_MONT_MOD_EXP3_H_		1

#include <spu_intrinsics.h>
#include "mpm_mont_mod_exp2.h"
#include "mpm_mul_inv.h"
#include "mpm_mod.h"


/* mpm_mont_mod_exp3
 * ----------------
 * Generic routine that computes modular exponentiation using Montgomery
 * modular multiplication. This routine computes:
 *
 *	c = (b ^ e) % m
 *
 * where b, e, and m are large multi-precision numbers of bsize, esize, 
 * and msize quadwords, respectively. The result, c, is of msize quadwords.
 *
 * This implementation uses a window-based modular exponentiation algorithm.
 */

static vector unsigned int _two_pow_n_mpm_mont_mod_exp3[2*MPM_MAX_SIZE+1] = {((vector unsigned int) { 0,0,0,1})};

static __inline void _mpm_mont_mod_exp3(vector unsigned int *c, const vector unsigned int *b, int bsize, const vector unsigned int *e, int esize, const vector unsigned int *m, int msize)
{
  vector unsigned int p, g[2], addend, no_carry;
  vector unsigned int a[MPM_MAX_SIZE];
  vector unsigned int u[MPM_MAX_SIZE];
  vector unsigned int b_mod_m[MPM_MAX_SIZE];
  vector unsigned int m_mod[2];

  /* Reduce the range of b to that of the modulus
   */
  _mpm_mod(b_mod_m, b, bsize, m, msize);

  /* Pre-compute the constant p, a, u used by the Montgomery
   * modular exponentiaion algorithm. 
   */
  m_mod[0] = spu_splats((unsigned int)0);
  m_mod[1] = m[msize-1];

  _mpm_mul_inv(g, _two_pow_n_mpm_mont_mod_exp3, m_mod, 2);

  /* p = -g */
  p = spu_xor(g[1], -1);
  no_carry = spu_xor(spu_cmpeq(g[1], 0), -1);

  addend = spu_xor(spu_or(spu_slqwbyte(no_carry, 12),
			  spu_or(spu_slqwbyte(no_carry, 8),
				 spu_slqwbyte(no_carry, 4))), -1);
  p = spu_sub(p, addend);
  _mpm_mod(a, _two_pow_n_mpm_mont_mod_exp3, msize+1, m, msize);
  _mpm_mod(u, _two_pow_n_mpm_mont_mod_exp3, 2*msize+1, m, msize);

  _mpm_mont_mod_exp2(c, b_mod_m, e, esize, m, msize, MPM_MOD_EXP_MAX_K, p, a, u);
}

#endif /* _MPM_MONT_MOD_EXP3_H_ */
