/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2007,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#ifndef _MPM_SWAP_ENDIAN_H_
#define _MPM_SWAP_ENDIAN_H_	1

#include <spu_intrinsics.h>

/* mpm_swap_endian
 * ---------------
 * Swap the endian-ness (ie. byte ordering) of the large number of <size>
 * quadwords pointed to by <a>. This routine converts little endian numbers
 * to big endian numbers and vice versa.
 */
static __inline void _mpm_swap_endian(vector unsigned int *a, int size)
{
  int i;
  vector unsigned int a0, a1;
  vector unsigned char byte_swap = ((vector unsigned char) {
					       0xF,0xE,0xD,0xC,
					       0xB,0xA,0x9,0x8,
					       0x7,0x6,0x5,0x4,
					       0x3,0x2,0x1,0x0});
  
  for (i=0; i<size/2; i++) {
    a0 = a[i];
    a1 = a[size-1-i];
    a[size-1-i] = spu_shuffle(a0, a0, byte_swap);
    a[i] = spu_shuffle(a1, a1, byte_swap);
  }
  if (size & 1) {
    a0 = a[i];
    a[i] = spu_shuffle(a0, a0, byte_swap);
  }
}

#endif /* _MPM_SWAP_ENDIAN_H_ */
